@@ -7,7 +7,7 @@ from harmony_model_checker.harmony.ops import *
 from harmony_model_checker.exception import *
 
 labelcnt = 0
-imported: Dict[str, Scope] = {}           # imported modules
+imported: Dict[str, Scope] = {}         # imported modules
 constants: Dict[str, str] = {}          # constants modified with -c
 used_constants = set()  # constants modified and used
 
@@ -83,12 +83,14 @@ class AST:
                 message='constant evaluation failed: %s %s' % (self, ctx.failure),
                 lexeme=lexeme,
                 filename=file,
+                line=line,
                 column=column
             )
         return ctx.pop()
 
     def compile(self, scope: Scope, code: Code, stmt):
-        if self.isConstant(scope):
+        # constant folding:  TODO
+        if False and self.isConstant(scope):
             code2 = Code(code)
             code2.modpush(code.curModule)
             self.gencode(scope, code2, stmt)
@@ -106,6 +108,7 @@ class AST:
             lexeme=lexeme,
             filename=file,
             # stmt=stmt,
+            line=line,
             column=column,
             message='Not a variable: %s' % str(self)
         )
@@ -118,6 +121,7 @@ class AST:
             lexeme=lexeme,
             filename=file,
             # stmt=stmt,
+            line=line,
             column=column,
             message='Cannot use in left-hand side expression: %s' % str(self)
         )
@@ -130,6 +134,7 @@ class AST:
             lexeme=lexeme,
             filename=file,
             # stmt=stmt,
+            line=line,
             column=column,
             message='Cannot take address of %s' % str(self)
         )
@@ -202,7 +207,7 @@ class ComprehensionAST(AST):
 
             global labelcnt
             startlabel = LabelValue(None, "$%d_start" % labelcnt)
-            endlabel = LabelValue(None, "$%d_end" % labelcnt)
+            endlabel = LabelValue(None, "$%d_end1" % labelcnt)
             labelcnt += 1
             code.nextLabel(startlabel)
             code.append(CutOp(var, var2), self.token, self.token, stmt=stmt)
@@ -293,7 +298,7 @@ class NameAST(AST):
             code.append(LoadVarOp(self.name), self.token, self.endtoken, stmt=stmt)
             if self.name[0] == "pre":
                 scope.uses_pre = True
-        elif t == "constant":
+        elif t in { "constant", "imported" }:
             (lexeme, file, line, column) = self.name
             code.append(PushOp(v), self.token, self.endtoken, stmt=stmt)
         else:
@@ -304,12 +309,13 @@ class NameAST(AST):
     # TODO.  How about local-const?
     def localVar(self, scope):
         (t, v) = scope.lookup(self.name)
-        assert t in {"constant", "local-var", "local-const", "global", "module"}
+        assert t in {"constant", "imported", "local-var", "local-const", "global", "module"}
         return self.name[0] if t in { "local-var", "local-const" } else None
 
     def ph1(self, scope, code, stmt):
+        # TODO: what if lexeme == "_"?
         (t, v) = scope.lookup(self.name)
-        if t == "constant":
+        if t in { "constant", "imported" }:
             (lexeme, file, line, column) = v
             code.append(PushOp((AddressValue(lexeme, []), file, line, column)), self.token, self.endtoken, stmt=stmt)
         elif t == "local-const":
@@ -318,19 +324,39 @@ class NameAST(AST):
             code.append(NaryOp(("Closure", file, line, column), 1), self.token, self.endtoken, stmt=stmt)
         elif t == "local-var":
             (lexeme, file, line, column) = v
-            # TODO: what if lexeme == "_"?
-            if lexeme != "_":
-                code.append(PushOp((AddressValue(PcValue(-2), [lexeme]), file, line, column)), self.token, self.endtoken, stmt=stmt)
+            code.append(PushOp((AddressValue(PcValue(-2), [lexeme]), file, line, column)), self.token, self.endtoken, stmt=stmt)
         else:
-            assert t == "global"
             (lexeme, file, line, column) = self.name
+            if t != "global":
+                raise HarmonyCompilerError(
+                    message="not a variable: " + lexeme,
+                    lexeme=lexeme,
+                    filename=file,
+                    line=line,
+                    column=column
+                )
             if scope.prefix is None:
                 code.append(PushOp((AddressValue(PcValue(-1), [lexeme]), file, line, column)), self.token, self.endtoken, stmt=stmt)
             else:
                 code.append(PushOp((AddressValue(PcValue(-1), [scope.prefix + '$' + lexeme]), file, line, column)), self.token, self.endtoken, stmt=stmt)
 
     def address(self, scope, code, stmt):
-        return self.ph1(scope, code, stmt)
+        # TODO: what if lexeme == "_"?
+        (t, v) = scope.lookup(self.name)
+        if t in { "constant", "imported" }:
+            (lexeme, file, line, column) = v
+            code.append(PushOp((AddressValue(lexeme, []), file, line, column)), self.token, self.endtoken, stmt=stmt)
+        elif t in { "local-const", "local-var" }:
+            (lexeme, file, line, column) = v
+            code.append(LoadVarOp(v), self.token, self.endtoken, stmt=stmt)
+            code.append(NaryOp(("Closure", file, line, column), 1), self.token, self.endtoken, stmt=stmt)
+        else:
+            assert t == "global"
+            (lexeme, file, line, column) = self.name
+            if scope.prefix == None:
+                code.append(PushOp((AddressValue(PcValue(-1), [lexeme]), file, line, column)), self.token, self.endtoken, stmt=stmt)
+            else:
+                code.append(PushOp((AddressValue(PcValue(-1), [scope.prefix + '$' + lexeme]), file, line, column)), self.token, self.endtoken, stmt=stmt)
 
     def ph2(self, scope, code, skip, start, stop, stmt):
         if skip > 0:
@@ -339,6 +365,8 @@ class NameAST(AST):
         (t, v) = scope.lookup(self.name)
         if t == "local-var":
             if self.name[0] == "_":
+                # Pop address and value
+                code.append(PopOp(), start, stop, stmt=stmt)
                 code.append(PopOp(), start, stop, stmt=stmt)
             else:
                 code.append(StoreVarOp(None, self.name[0]), start, stop, stmt=stmt)
@@ -351,7 +379,7 @@ class NameAST(AST):
         (t, v) = scope.lookup(self.name)
         if t in {"local-var", "local-const", "global", "module"}:
             return False
-        elif t == "constant":
+        elif t in { "constant", "imported" }:
             return not isinstance(v[0], LabelValue)
         else:
             assert False, (t, v, self.name)
@@ -439,6 +467,7 @@ class TupleAST(AST):
             message="Cannot index into tuple in assignment",
             lexeme=lexeme,
             filename=file,
+            line=line,
             column=column
         )
 
@@ -568,7 +597,7 @@ class NaryAST(AST):
 
     def isConstant(self, scope):
         (op, file, line, column) = self.op
-        if op in {"atLabel", "choose", "contexts", "countLabel", "get_context"}:
+        if op in {"choose", "contexts", "get_context", "get_ident"}:
             return False
         return all(x.isConstant(scope) for x in self.args)
 
@@ -579,7 +608,7 @@ class NaryAST(AST):
         if op == "and" or op == "or":
             self.args[0].compile(scope, code, stmt)
             lastlabel = LabelValue(None, "$%d_last" % labelcnt)
-            endlabel = LabelValue(None, "$%d_end" % labelcnt)
+            endlabel = LabelValue(None, "$%d_end2" % labelcnt)
             labelcnt += 1
             for i in range(1, n):
                 code.append(JumpCondOp(op == "or", lastlabel), self.token, self.endtoken, stmt=stmt)
@@ -592,7 +621,7 @@ class NaryAST(AST):
             assert n == 2, n
             self.args[0].compile(scope, code, stmt)
             truelabel = LabelValue(None, "$%d_true" % labelcnt)
-            endlabel = LabelValue(None, "$%d_end" % labelcnt)
+            endlabel = LabelValue(None, "$%d_end3" % labelcnt)
             labelcnt += 1
             code.append(JumpCondOp(False, truelabel), self.token, self.endtoken, stmt=stmt)
             self.args[1].compile(scope, code, stmt)
@@ -606,7 +635,7 @@ class NaryAST(AST):
             cond = self.args[1].args[0] if negate else self.args[1]
             cond.compile(scope, code, stmt)
             elselabel = LabelValue(None, "$%d_else" % labelcnt)
-            endlabel = LabelValue(None, "$%d_end" % labelcnt)
+            endlabel = LabelValue(None, "$%d_end4" % labelcnt)
             labelcnt += 1
             code.append(JumpCondOp(negate, elselabel), self.token, self.endtoken, stmt=stmt)
             self.args[0].compile(scope, code, stmt)  # "if" expr
@@ -630,6 +659,47 @@ class NaryAST(AST):
     def accept_visitor(self, visitor, *args, **kwargs):
         return visitor.visit_nary(self, *args, **kwargs)
 
+class BinAST(AST):
+    def __init__(self, endtoken, token, ops, args):
+        AST.__init__(self, endtoken, token, False)
+        self.ops = ops
+        self.args = args
+        assert len(ops) == len(args) - 1
+        assert all(isinstance(x, AST) for x in args), args
+
+    def gencode(self, scope, code, stmt):
+        n = len(self.args)
+        self.args[0].compile(scope, code, stmt)
+        (lexeme, file, line, column) = self.ops[0]
+        T = ("__cmp__" + str(len(code.labeled_ops)), file, line, column)
+        endlabel = LabelValue(None, "cmp$%d"%len(code.labeled_ops))
+        for i in range(1, n - 1):
+            self.args[i].compile(scope, code, stmt)
+            code.append(DupOp(), self.token, self.endtoken, stmt=stmt)
+            code.append(StoreVarOp(T), self.token, self.endtoken, stmt=stmt)
+            code.append(NaryOp(self.ops[i - 1], 2), self.token, self.endtoken, stmt=stmt)
+            code.append(DupOp(), self.token, self.endtoken, stmt=stmt)
+            code.append(JumpCondOp(False, endlabel), self.token, self.endtoken, stmt=stmt)
+            code.append(PopOp(), self.token, self.endtoken, stmt=stmt)
+            code.append(LoadVarOp(T), self.token, self.endtoken, stmt=stmt)
+        self.args[n - 1].compile(scope, code, stmt)
+        code.append(NaryOp(self.ops[n - 2], 2), self.token, self.endtoken, stmt=stmt)
+        code.nextLabel(endlabel)
+        if n > 2:
+            code.append(DelVarOp(T), self.token, self.endtoken, stmt=stmt)     # TODO: is this necessary???
+
+    def getLabels(self):
+        labels = [x.getLabels() for x in self.args]
+        return functools.reduce(lambda x, y: x | y, labels)
+
+    def accept_visitor(self, visitor, *args, **kwargs):
+        return visitor.visit_cmp(self, *args, **kwargs)
+
+    def __repr__(self):
+        return "BinOp(" + str(self.ops) + ", " + str(self.args) + ")"
+
+    def isConstant(self, scope):
+        return all(x.isConstant(scope) for x in self.args)
 
 class CmpAST(AST):
     def __init__(self, endtoken, token, ops, args):
@@ -692,18 +762,30 @@ class ApplyAST(AST):
             # See if it's of the form "module.constant":
             if t == "module":
                 if isinstance(self.arg, ConstantAST) and isinstance(self.arg.const[0], str):
-                    (t2, v2) = v.lookup(self.arg.const)
-                    assert t2 == "constant"
+                    tv = v.find(self.arg.const)
+                    if not tv:
+                        (lexeme, file, line, column) = self.arg.const
+                        raise HarmonyCompilerError(
+                            message="can't find %s in module %s" % (lexeme, self.method.name[0]),
+                            lexeme=lexeme,
+                            filename=file,
+                            line=line,
+                            column=column
+                        )
+
+                    (t2, v2) = tv
+                    assert t2 in { "constant", "imported" }, (t2, v2)
                     code.append(PushOp(v2), self.token, self.endtoken, stmt=stmt)
                     return True
                 raise HarmonyCompilerError(
                     message="can only look up a constant in a module %s, not %s" % (self.method.name, self.arg),
                     lexeme=lexeme,
                     filename=file,
+                    line=line,
                     column=column
                 )
 
-            if t == "constant":
+            if t in { "constant", "imported" }:
                 code.append(PushOp(v), self.token, self.endtoken, stmt=stmt)
                 self.arg.compile(scope, code, stmt)
                 code.append(NaryOp(("Closure", file, line, column), 2), self.token, self.endtoken, stmt=stmt)
@@ -715,6 +797,7 @@ class ApplyAST(AST):
                         message="can't apply to _",
                         lexeme=lexeme,
                         filename=file,
+                        line=line,
                         column=column
                     )
                 self.method.compile(scope, code, stmt)  # LoadVar
@@ -766,13 +849,14 @@ class ApplyAST(AST):
         # TODO.  Can same be done with lambda?
         if isinstance(self.method, NameAST):
             (t, v) = scope.lookup(self.method.name)
-            if t == "constant" and isinstance(v[0], LabelValue):
+            if t in { "constant", "imported" } and isinstance(v[0], LabelValue):
                 self.arg.compile(scope, code, stmt)
                 code.append(ApplyOp(v), self.token, self.endtoken, stmt=stmt)
                 return
         if self.varcompile(scope, code, stmt):
-            (lexeme, file, line, column) = self.token
-            code.append(NaryOp(("Closure", file, line, column), 2), self.token, self.endtoken, stmt=stmt)
+            # (lexeme, file, line, column) = self.token
+            # code.append(NaryOp(("Closure", file, line, column), 2), self.token, self.endtoken, stmt=stmt)
+            pass
         else:
             code.append(LoadOp(None, self.token, None), self.token, self.endtoken, stmt=stmt)
 
@@ -787,11 +871,12 @@ class ApplyAST(AST):
             (t, v) = scope.lookup(self.method.name)
             if t == "module" and isinstance(self.arg, ConstantAST) and isinstance(self.arg.const[0], str):
                 (t2, v2) = v.lookup(self.arg.const)
-                assert t2 == "constant"
+                assert t2 in { "constant", "imported" }
                 raise HarmonyCompilerError(
                     message="Cannot assign to constant %s %s" % (self.method.name, self.arg.const),
                     lexeme=lexeme,
                     filename=file,
+                    line=line,
                     column=column
                 )
 
@@ -813,11 +898,12 @@ class ApplyAST(AST):
             (t, v) = scope.lookup(self.method.name)
             if t == "module" and isinstance(self.arg, ConstantAST) and isinstance(self.arg.const[0], str):
                 (t2, v2) = v.lookup(self.arg.const)
-                assert t2 == "constant"
+                assert t2 in { "constant", "imported" }
                 raise HarmonyCompilerError(
                     message="Cannot assign to constant %s %s" % (self.method.name, self.arg.const),
                     lexeme=lexeme,
                     filename=file,
+                    line=line,
                     column=column
                 )
 
@@ -897,7 +983,7 @@ class AssignmentAST(AST):
 
         # Compute the addresses of lhs expressions
         for lvs in self.lhslist:
-            # handled separately for better assembly code readability
+            # NameAST handled separately for better assembly code readability
             if not isinstance(lvs, NameAST):
                 lvs.ph1(scope, code, stmt)
 
@@ -908,7 +994,10 @@ class AssignmentAST(AST):
         for i in range(len(self.lhslist) - 1):
             code.append(DupOp(), self.ops[i], self.ops[i], stmt=stmt)
 
-        # Now assign to the left-hand side in reverse order
+        # Now assign to the left-hand side in reverse order.  We assign
+        # in reverse order because the right-hand-side was evaluated in
+        # forward order and the values appear in that way on the stack.
+        # TODO.  Could change this by reversing the semantics of Split.
         skip = len(self.lhslist)
         for lvs in reversed(self.lhslist):
             skip -= 1
@@ -919,27 +1008,26 @@ class AssignmentAST(AST):
                     raise HarmonyCompilerError(
                         lexeme=lexeme,
                         filename=file,
+                        line=line,
                         column=column,
                         message='Cannot assign to module %s' % str(lvs.name),
                     )
-                if t in {"constant", "local-const"}:
+                if t in {"constant", "imported", "local-const"}:
                     raise HarmonyCompilerError(
                         lexeme=lexeme,
                         filename=file,
+                        line=line,
                         column=column,
                         message='Cannot assign to constant %s' % str(lvs.name),
                     )
                 assert t in {"local-var", "global"}, (t, lvs.name)
                 if v[0] == "_":
+                    assert False
                     code.append(PopOp(), lvs.token, self.ops[skip], stmt=stmt)
                 else:
                     st = StoreOp(lvs.name, lvs.name, scope.prefix) if t == "global" else StoreVarOp(lvs.name)
                     code.append(st, lvs.token, self.ops[skip], stmt=stmt)
             else:
-                # if skip > 0:
-                #     code.append(MoveOp(skip + 2), self.token, self.endtoken, stmt=stmt)
-                #     code.append(MoveOp(2), self.token, self.endtoken, stmt=stmt)
-                # code.append(StoreOp(None, self.token, None), lvs.token, self.ops[skip], stmt=stmt)
                 lvs.ph2(scope, code, skip, lvs.token, self.ops[skip], stmt)
 
         if self.atomically:
@@ -978,13 +1066,15 @@ class AuxAssignmentAST(AST):
                 raise HarmonyCompilerError(
                     filename=file,
                     lexeme=lexeme,
+                    line=line,
                     column=column,
                     message='Cannot operate on module %s' % str(lv.name),
                 )
-            if t in {"constant", "local-const"}:
+            if t in {"constant", "imported", "local-const"}:
                 raise HarmonyCompilerError(
                     filename=file,
                     lexeme=lexeme,
+                    line=line,
                     column=column,
                     message='Cannot operate on constant %s' % str(lv.name),
                 )
@@ -1027,9 +1117,11 @@ class DelAST(AST):
             code.append(AtomicIncOp(False), self.token, self.endtoken, stmt=stmt)
         lvar = self.lv.localVar(scope)
         if isinstance(self.lv, NameAST):
+            # TODO.  Is DelVarOp necessary??
             op = DelOp(self.lv.name, scope.prefix) if lvar is None else DelVarOp(self.lv.name)
         else:
             self.lv.ph1(scope, code, stmt)
+            # TODO.  Is DelVarOp necessary??
             op = DelOp(None, None) if lvar is None else DelVarOp(None, lvar)
         code.append(op, self.token, self.endtoken, stmt=stmt)
         if self.atomically:
@@ -1121,6 +1213,7 @@ class AddressAST(AST):
                 raise HarmonyCompilerError(
                     filename=file,
                     lexeme=lexeme,
+                    line=line,
                     column=column,
                     message="Can't take address of imported %s" % str(lv),
                 )
@@ -1139,6 +1232,7 @@ class AddressAST(AST):
             raise HarmonyCompilerError(
                 filename=file,
                 lexeme=lexeme,
+                line=line,
                 column=column,
                 message="Can't take address of %s" % str(lv),
             )
@@ -1175,7 +1269,7 @@ class BlockAST(AST):
     def __init__(self, endtoken, token, atomically, b, colon):
         AST.__init__(self, endtoken, token, atomically)
         assert len(b) > 0
-        self.b = b
+        self.b = b      # list of statements
         self.colon = colon
 
     def __repr__(self):
@@ -1184,9 +1278,21 @@ class BlockAST(AST):
     def compile(self, scope, code, stmt):
         stmt = self.range(self.token, self.colon)
         ns = Scope(scope)
+
+        # Labels are global
         for s in self.b:
-            for ((lexeme, file, line, column), lb) in s.getLabels():
+            for (token, lb) in s.getLabels():
+                (lexeme, file, line, column) = token
+                if False and ns.find(token):
+                    raise HarmonyCompilerError(
+                        filename=file,
+                        lexeme=lexeme,
+                        line=line,
+                        column=column,
+                        message="%s: Parse error: duplicate definition" % lexeme
+                    )
                 ns.names[lexeme] = ("constant", (lb, file, line, column))
+
         if self.atomically:
             code.append(AtomicIncOp(False), self.token, self.endtoken, stmt=stmt)
         for s in self.b:
@@ -1224,7 +1330,7 @@ class IfAST(AST):
         label = labelcnt
         labelcnt += 1
         sublabel = 0
-        endlabel = LabelValue(None, "$%d_end" % label)
+        endlabel = LabelValue(None, "$%d_end5" % label)
         if self.atomically:
             code.append(AtomicIncOp(False), self.token, self.endtoken, stmt=stmt)
         last = len(self.alts) - 1
@@ -1282,7 +1388,7 @@ class WhileAST(AST):
         cond = self.cond.args[0] if negate else self.cond
         global labelcnt
         startlabel = LabelValue(None, "$%d_start" % labelcnt)
-        endlabel = LabelValue(None, "$%d_end" % labelcnt)
+        endlabel = LabelValue(None, "$%d_end6" % labelcnt)
         labelcnt += 1
         code.nextLabel(startlabel)
         cond.compile(scope, code, stmt)
@@ -1348,7 +1454,8 @@ class InvariantAST(AST):
         code.append(JumpOp(endlabel, reason="jump over invariant definition"), self.token, self.token, stmt=stmt)
         code.nextLabel(startlabel)
         (_, file, line, column) = self.token
-        args = [ ("pre", file, line, column), ("post", file, line, column) ]
+        # args = [ ("pre", file, line, column), ("post", file, line, column) ]
+        args = []
         code.append(FrameOp(self.token, args), self.token, self.endtoken, stmt=stmt)
 
         ns = Scope(scope)
@@ -1577,6 +1684,9 @@ class AtomicAST(AST):
     def compile(self, scope, code, stmt):
         stmt = self.range(self.atomically, self.colon)
         code.append(AtomicIncOp(False), self.token, self.endtoken, stmt=stmt)
+        if isinstance(self.stat, BlockAST) and len(self.stat.b) == 1 and isinstance(self.stat.b[0], LocationAST) and isinstance(self.stat.b[0].ast, LetWhenAST):
+            (lexeme, file, line, column) = self.token
+            print('Warning: you probably want to do "atomically when" rather than "atomically: when" in %s, line %d'%(file, line))
         self.stat.compile(scope, code, stmt)
         code.append(AtomicDecOp(), self.token, self.endtoken, stmt=stmt)
 
@@ -1764,6 +1874,7 @@ class CallAST(AST):
 
     def compile(self, scope, code, stmt):
         stmt = self.stmt()
+        # TODO.  What does the following if statement do again?
         if not self.expr.isConstant(scope):
             if self.atomically:
                 code.append(AtomicIncOp(False), self.token, self.endtoken, stmt=stmt)
@@ -1780,21 +1891,19 @@ class CallAST(AST):
 
 
 class SpawnAST(AST):
-    def __init__(self, endtoken, token, atomically, method, arg, this, eternal):
+    def __init__(self, endtoken, token, atomically, expr, this, eternal):
         AST.__init__(self, endtoken, token, atomically)
-        self.method = method
-        self.arg = arg
+        self.expr = expr
         self.this = this
         self.eternal = eternal
 
     def __repr__(self):
-        return "Spawn(" + str(self.method) + ", " + str(self.arg) + ", " + str(self.this) + ", " + str(
+        return "Spawn(" + str(self.expr) + ", " + str(self.this) + ", " + str(
             self.eternal) + ")"
 
     def compile(self, scope, code, stmt):
         stmt = self.stmt()
-        self.method.compile(scope, code, stmt)
-        self.arg.compile(scope, code, stmt)
+        self.expr.ph1(scope, code, stmt)
         if self.this is None:
             code.append(PushOp((emptydict, None, None, None)), self.token, self.endtoken, stmt=stmt)
         else:
@@ -1803,30 +1912,33 @@ class SpawnAST(AST):
 
     # TODO.  Deal with 'this'
     def getLabels(self):
-        return self.method.getLabels() | self.arg.getLabels()
+        return self.expr.getLabels()
 
     def accept_visitor(self, visitor, *args, **kwargs):
         return visitor.visit_spawn(self, *args, **kwargs)
 
 
 class TrapAST(AST):
-    def __init__(self, endtoken, token, atomically, method, arg):
+    def __init__(self, endtoken, token, atomically, expr):
         AST.__init__(self, endtoken, token, atomically)
-        self.method = method
-        self.arg = arg
+        self.expr = expr
 
     def __repr__(self):
-        return "Trap(" + str(self.method) + ", " + str(self.arg) + ")"
+        return "Trap(" + str(self.expr) + ")"
 
     def compile(self, scope, code, stmt):
         stmt = self.stmt()
-        # TODO.  These should be swapped
-        self.arg.compile(scope, code, stmt)
-        self.method.compile(scope, code, stmt)
+        if isinstance(self.expr, ApplyAST):      # backward compatibility
+            self.expr.method.ph1(scope, code, stmt)
+            self.expr.arg.compile(scope, code, stmt)
+            (_, file, line, column) = self.token
+            code.append(NaryOp(("AddArg", file, line, column), 2), self.token, self.endtoken, stmt=stmt)
+        else:
+            self.expr.compile(scope, code, stmt)
         code.append(TrapOp(), self.token, self.endtoken, stmt=stmt)
 
     def getLabels(self):
-        return self.method.getLabels() | self.arg.getLabels()
+        return self.expr.getLabels()
 
     def accept_visitor(self, visitor, *args, **kwargs):
         return visitor.visit_trap(self, *args, **kwargs)
@@ -1899,18 +2011,26 @@ class FromAST(AST):
         if self.items == []:  # from module import *
             for (item, (t, v)) in names.items():
                 if t == "constant":
-                    scope.set(item, (t, v))
+                    scope.tryset((item, file, line, column), ("imported", v))
         else:
-            for (lexeme, file, line, column) in self.items:
+            for token in self.items:
+                (lexeme, file, line, column) = token
                 if lexeme not in names:
                     raise HarmonyCompilerError(
                         filename=file,
                         lexeme=lexeme,
                         message="%s line %s: can't import %s from %s" % (file, line, lexeme, self.module[0]),
+                        line=line,
                         column=column)
                 (t, v) = names[lexeme]
-                assert t == "constant", (lexeme, t, v)
-                scope.set(lexeme, (t, v))
+                if t != "constant":
+                    raise HarmonyCompilerError(
+                        filename=file,
+                        lexeme=lexeme,
+                        message="%s line %s: can't import %s from %s: not a constant" % (file, line, lexeme, self.module[0]),
+                        line=line,
+                        column=column)
+                scope.tryset(token, ("imported", v))
 
     def getImports(self):
         return [self.module]
@@ -1977,6 +2097,36 @@ class LabelStatAST(AST):
         return visitor.visit_label_stat(self, *args, **kwargs)
 
 
+class GlobalAST(AST):
+    def __init__(self, endtoken, token, atomically, vars):
+        AST.__init__(self, endtoken, token, atomically)
+        self.vars = vars
+
+    def __repr__(self):
+        return "Global(" + str(self.vars) + ")"
+
+    def compile(self, scope, code, stmt):
+        for var in self.vars:
+            assert isinstance(var, NameAST)
+            (t, v) = scope.lookup(var.name)
+            lexeme, file, line, column = var.name
+            if t != "global":
+                raise HarmonyCompilerError(
+                    filename=file,
+                    lexeme=lexeme,
+                    line=line,
+                    column=column,
+                    message="%s: Parse error: shadows prior definition" % lexeme
+                )
+            # scope.set(lexeme, ("global", (lexeme, file, line, column)))
+
+    def getLabels(self):
+        return set()
+
+    def accept_visitor(self, visitor, *args, **kwargs):
+        return visitor.visit_sequential(self, *args, **kwargs)
+
+
 class SequentialAST(AST):
     def __init__(self, endtoken, token, atomically, vars):
         AST.__init__(self, endtoken, token, atomically)
@@ -1986,6 +2136,18 @@ class SequentialAST(AST):
         return "Sequential(" + str(self.vars) + ")"
 
     def compile(self, scope, code, stmt):
+        for var in self.vars:
+            assert isinstance(var, NameAST)
+            (t, v) = scope.lookup(var.name)
+            lexeme, file, line, column = var.name
+            if t != "global":
+                raise HarmonyCompilerError(
+                    filename=file,
+                    lexeme=lexeme,
+                    line=line,
+                    column=column,
+                    message="%s: Parse error: shadows prior definition" % lexeme
+                )
         stmt = self.stmt()
         for lv in self.vars:
             lv.ph1(scope, code, stmt)
@@ -2031,10 +2193,12 @@ class ConstAST(AST):
     def set(self, scope, const, v):
         if isinstance(const, tuple):
             (lexeme, file, line, column) = const
-            if lexeme in scope.names:
+            tv = scope.find(const)
+            if tv:
                 raise HarmonyCompilerError(
                     filename=file,
                     lexeme=lexeme,
+                    line=line,
                     column=column,
                     message="%s: Parse error: already defined" % lexeme
                 )
@@ -2058,12 +2222,15 @@ class ConstAST(AST):
             raise HarmonyCompilerError(
                 filename=file,
                 lexeme=lexeme,
+                line=line,
                 column=column,
                 message="%s: Parse error: expression not a constant %s" % (self.const, self.expr),
             )
         code2 = Code(code)
         code2.modpush(code.curModule)
         self.expr.compile(scope, code2, stmt)
+        code2.append(ContinueOp(), self.token, self.endtoken, stmt=stmt)      # Hack: get endlabels evaluated
+        code2.link()
         state = State(code2, scope.labels)
         ctx = ContextValue(("__const__", None, None, None), 0, emptytuple, emptydict)
         ctx.atomic = 1
