@@ -13,6 +13,11 @@ var coderow = document.getElementById("coderow");
 var container = document.getElementById('table-scroll');
 var currOffset = 0;
 var currCloc = null;
+var colors = [ "green", "coral", "blueviolet" ];
+
+// printing contexts
+var contexts = {};
+var ctxGen = 0;
 
 function json_string_list(obj) {
   var result = "";
@@ -25,6 +30,17 @@ function json_string_list(obj) {
   return "[ " + result + " ]";
 }
 
+function json_string_tuple(obj) {
+  var result = "";
+  for (var i = 0; i < obj.length; i++) {
+    if (result != "") {
+      result += ", ";
+    }
+    result += json_string(obj[i]);
+  }
+  return "(" + result + ")";
+}
+
 function json_string_set(obj) {
   var result = "";
   for (var i = 0; i < obj.length; i++) {
@@ -101,8 +117,30 @@ function json_string_address(func, args) {
 }
 
 function json_string_context(obj) {
-  var pc = json_string(obj.pc);
-  return "CTX(" + pc + ")";
+  // TODO.  Is JSON.stringify deterministic (same context --> same string)?
+  var key = JSON.stringify(obj);
+  if (!(key in contexts)) {
+    contexts[key] = ++ctxGen;
+  }
+  var output = '<span title="CONTEXT\n';
+  if ("id" in obj) {
+    output += "id: " + json_string(obj["id"]) + "\n";
+  }
+  output += "pc: " + json_string(obj["pc"]) + "\n";
+  output += "vars: " + stringify_vars(obj["vars"]) + "\n";
+  var atomic = "atomic" in obj && obj["atomic"]["value"] == "True";
+  var stopped = "stopped" in obj && obj["stopped"]["value"] == "True";
+  if (atomic || stopped) {
+    output += "mode:"
+    if (atomic) output += " atomic";
+    if (stopped) output += " stopped";
+    output += "\n";
+  }
+  output += "sp: " + json_string(obj["sp"]);
+  return output + '" style="color:blue">C' + contexts[key] + '</span>';
+
+  // var pc = json_string(obj.pc);
+  // return "CTX(" + pc + ")";
 }
 
 function json_string(obj) {
@@ -150,25 +188,34 @@ function getCode(pc) {
 //  return noloc;
 }
 
-function drawTimeLine(mes) {
+function drawTimeLine(mes, mis) {
   var c = mes.canvas.getContext("2d");
   c.beginPath();
   c.clearRect(0, 0, mes.canvas.width, mes.canvas.height);
   var t = mes.startTime;
-  var yboxes = Math.floor((mes.nsteps + timeWidth - 1) / timeWidth);
   var nsteps = mes.nsteps;
+
+  // Compute the total number of rows
+  var yboxes = Math.ceil(mes.nsteps / timeWidth);
+
   for (var y = 0; y < yboxes; y++) {
     var xboxes = nsteps > timeWidth ? timeWidth : nsteps;
     for (var x = 0; x < xboxes; x++) {
-      c.fillStyle = t < currentTime ? "orange" : "white";
+      if (t >= currentTime) {
+        c.fillStyle = "white";
+      }
+      else {
+        c.fillStyle = colors[mis[t].trace.length % colors.length];
+      }
+      // c.fillStyle = t < currentTime ? "orange" : "white";
       c.fillRect(x * boxSize, y * boxSize, boxSize, boxSize);
       c.rect(x * boxSize, y * boxSize, boxSize, boxSize);
-      c.stroke();
       t += 1;
     }
     nsteps -= xboxes;
   }
-  mes.nextstep.innerHTML = "";
+  c.stroke();
+  // mes.nextstep.innerHTML = "";
   return t;
 }
 
@@ -181,6 +228,10 @@ function updateStatus(mes) {
       }
       else if (c.hasOwnProperty("next")) {
         switch (c.next.type) {
+        case "Continue":
+          var loc = getCode(c.pc);
+          mes.nextstep.innerHTML = "stopped in " + '<a href="' + loc.file + '">' + loc.module + "</a>:" + loc.line;
+          break;
         case "Store":
           mes.nextstep.innerHTML = "about to store " + json_string(c.next.value) + " in variable " + c.next.var;
           break;
@@ -218,7 +269,7 @@ function currentMegaStep() {
 function stringify_vars(obj) {
   var result = "";
   for (var k in obj) {
-    if (k == "result" && obj[k].type == "address" && !("func" in obj)) {
+    if (false && k == "result" && obj[k].type == "address" && !("func" in obj)) {
       continue;
     }
     if (result != "") {
@@ -253,6 +304,23 @@ function convert_vars(obj) {
   return result;
 }
 
+function method_call(m, arg) {
+  var result = "";
+  if (m.type == "atom") {
+    result += m.value;
+  }
+  else {
+    result += json_string(m);
+  }
+  if (arg.type == "list") {
+    result += json_string_tuple(arg.value);
+  }
+  else {
+    result += "(" + json_string(arg) + ")";
+  }
+  return result;
+}
+
 function stackTrace(tid, trace, failure) {
   var table = threads[tid].tracetable;
   table.innerHTML = "";
@@ -265,23 +333,29 @@ function stackTrace(tid, trace, failure) {
     var row = table.insertRow();
 
     var mcell = row.insertCell();
-    mcell.innerHTML = trace[i].method;
+    // mcell.innerHTML = trace[i].method;
+    mcell.innerHTML = method_call(trace[i].method_name, trace[i].method_arg);
     switch (trace[i].calltype) {
     case "process":
-        mcell.style.color = "blue";
+        // mcell.style.color = "blue";
+        mcell.style.color = colors[(i + 1) % colors.length];
         break;
     case "normal":
-        mcell.style.color = "black";
+        // mcell.style.color = "black";
+        mcell.style.color = colors[(i + 1) % colors.length];
         break;
     case "interrupt":
-        mcell.style.color = "orange";
+        // mcell.style.color = "orange";
+        mcell.style.color = colors[(i + 1) % colors.length];
         break;
     default:
         mcell.style.color = "red";
     }
 
     var vcell = row.insertCell();
-    var vtext = document.createTextNode(stringify_vars(trace[i].vars));
+    // var vtext = document.createTextNode(stringify_vars(trace[i].vars));
+    var vtext = document.createElement("div");
+    vtext.innerHTML = stringify_vars(trace[i].vars);
     vcell.appendChild(vtext);
   }
   if (false && failure != null) {
@@ -326,31 +400,53 @@ function handleKeyPress(e) {
       run_microsteps();
       break;
     case 'ArrowUp':
-      var mesidx = currentMegaStep();
-      var mes = megasteps[mesidx];
-      if (currentTime == mes.startTime && mesidx > 0) {
-          mes = megasteps[mesidx - 1];
+      if (currentTime > 0) {
+        currentTime--;
+        var mesidx = currentMegaStep();
+        var mes = megasteps[mesidx];
+        var mis = microsteps[currentTime];
+        var tl = mis.trace.length;
+        var endTime = mes.startTime + mes.nsteps;
+        while (currentTime > mes.startTime) {
+          mis = microsteps[currentTime - 1];
+          tl2 = mis.trace.length;
+          if (tl2 < tl) {
+            break;
+          }
+          currentTime--;
+        }
+        run_microsteps();
       }
-      currentTime = mes.startTime;
-      run_microsteps();
       break;
     case 'ArrowDown':
-      var mesidx = currentMegaStep();
-      var mes = megasteps[mesidx];
-      currentTime = mes.startTime + mes.nsteps;
-      if (currentTime > totalTime) {
-        currentTime = totalTime;
+      if (currentTime < totalTime) {
+        var mesidx = currentMegaStep();
+        var mes = megasteps[mesidx];
+        var mis = microsteps[currentTime];
+        var tl = mis.trace.length;
+        var endTime = mes.startTime + mes.nsteps;
+        while (++currentTime < endTime) {
+          mis = microsteps[currentTime];
+          tl2 = mis.trace.length;
+          if (tl2 < tl) {
+            break;
+          }
+        }
+        run_microsteps();
       }
-      run_microsteps();
       break;
     case 'Enter':
       if (currentTime < totalTime) {
+        var mesidx = currentMegaStep();
         var cloc = getCode(microsteps[currentTime].pc);
         while (++currentTime < totalTime) {
           var nloc = getCode(microsteps[currentTime].pc);
           if (nloc.file != cloc.file || nloc.line != cloc.line || nloc.code != cloc.code) {
             break;
           }
+          if (currentMegaStep() != mesidx) {
+            break;
+          }
         }
         run_microsteps();
       }
@@ -382,7 +478,7 @@ function init_microstep(masidx, misidx) {
     // invfails: misidx == mas.microsteps.length - 1 ? mas.invfails : [],
     contexts: mas.contexts,
     hvm: mis.code,
-    explain: mis.explain
+    explain2: mis.explain2
   };
   if (misidx != 0) {
     previous = microsteps[t-1];
@@ -514,6 +610,10 @@ function init_microstep(masidx, misidx) {
     microsteps[t].trace[trl - 1].vars = mis.local;
   }
 
+  // if (mis.pc == "963") {
+  //   alert(JSON.stringify(convert_vars(mis.shared)))
+  // }
+
   if (mis.hasOwnProperty("shared")) {
     microsteps[t].shared = convert_vars(mis.shared);
   }
@@ -601,12 +701,46 @@ function escapeHTML(s) {
      .replace(/'/g, "&#039;");
 }
 
+function explain_expand(e) {
+    var s = e.text;
+    var result = "";
+    var arg = 0;
+    for (var i = 0; i < s.length; i++) {
+      if (s[i] == '#') {
+        i++;
+        if (i == s.length) {
+          break;
+        }
+        if (s[i] == '#') {
+          result += '#';
+        }
+        else if (s[i] == '+') {
+          result += json_string(e.args[arg++]);
+        }
+        else if (s[i] == '@') {
+          var addr = json_string(e.args[arg++]);
+          result += addr.slice(1);
+        }
+        else {
+          result += json_string(e.args[parseInt(s[i])]);
+        }
+      }
+      else {
+        result += s[i];
+      }
+    }
+    return result;
+}
+
 function run_microstep(t) {
   var mis = microsteps[t];
   var mes = megasteps[mis.mesidx];
   var mesrow = mestable.rows[mis.mesidx];
   mesrow.cells[3].innerHTML = mis.npc;
 
+  // if (mis.pc == 963) {
+  //   alert(JSON.stringify(mis.shared));
+  // }
   for (var i = 0; i < vardir.length; i++) {
     mesrow.cells[i + 4].innerHTML = get_shared(mis.shared, vardir[i]);
   }
@@ -669,13 +803,13 @@ function run_microstep(t) {
   else if (t+1 < microsteps.length) {
     var nmis = microsteps[t+1];
     if (nmis.tid == mis.tid) {
-        mes.nextstep.innerHTML = "next: " + nmis.explain;
+        mes.nextstep.innerHTML = "next2: " + explain_expand(nmis.explain2);
     }
     else {
         updateStatus(mes);
     }
 
-    // hvmrow.innerHTML = "T" + nmis.tid + "/" + nmis.pc + ": " + nmis.hvm + " (" + nmis.explain + ")"
+    // hvmrow.innerHTML = "T" + nmis.tid + "/" + nmis.pc + ": " + nmis.hvm + " (" + explain_expand(nmis.explain2) + ")"
     currCloc = document.getElementById('C' + nmis.pc);
     currOffset = document.getElementById('P' + nmis.pc);
   }
@@ -708,48 +842,53 @@ function run_microsteps() {
   }
 
   for (var i = 0; i < nmegasteps; i++) {
-    drawTimeLine(megasteps[i]);
+    megasteps[i].nextstep.innerHTML = "";
   }
   for (var t = 0; t < currentTime; t++) {
     run_microstep(t);
   }
-  if (currentTime < microsteps.length && (currentTime == 0 ||
-            microsteps[currentTime - 1].tid != microsteps[currentTime].tid)) {
+  for (var i = 0; i < nmegasteps; i++) {
+    drawTimeLine(megasteps[i], microsteps);
+  }
+  if (currentTime < microsteps.length /* && (currentTime == 0 ||
+           microsteps[currentTime - 1].tid != microsteps[currentTime].tid) */) {
     var mis = microsteps[currentTime];
     var mes = megasteps[mis.mesidx];
-    mes.nextstep.innerHTML = "next: " + mis.explain;
+    mes.nextstep.innerHTML = "next: " + explain_expand(mis.explain2);
     var mesrow = mestable.rows[mis.mesidx];
     mesrow.cells[3].innerHTML = mis.pc;
 
-    if (mis.mesidx > 0) {
+    if (false && mis.mesidx > 0) {
       var oldrow = mestable.rows[mis.mesidx - 1];
       for (var i = 0; i < vardir.length; i++) {
         mesrow.cells[i + 4].innerHTML = oldrow.cells[i + 4].innerHTML;
       }
     }
   }
-  container.scrollTop = currOffset.offsetTop;
+  container.scrollTop = currOffset.offsetTop - 50;
 
   if (currCloc != null) {
     currCloc.style.color = "red";
   }
 
-  var curmes = currentTime < totalTime ? microsteps[currentTime].mesidx : -1;
+  var curmes = currentTime > 0 ? microsteps[currentTime - 1].mesidx : -1;
   for (var mes = 0; mes < nmegasteps; mes++) {
     var row = document.getElementById("mes" + mes);
     if (mes == curmes) {
-      row.style = 'background-color: #A5FF33;';
+      // row.style = 'background-color: #A5FF33;';
+      row.style = 'background-color: #DFFF00;';
     }
     else {
       row.style = 'background-color: white;';
     }
   }
 
-  var curtid = currentTime < totalTime ? microsteps[currentTime].tid : -1;
+  var curtid = currentTime > 0 ? microsteps[currentTime - 1].tid : -1;
   for (var tid = 0; tid < nthreads; tid++) {
     var row = document.getElementById("thread" + tid);
     if (tid == curtid) {
-      row.style = 'background-color: #A5FF33;';
+      // row.style = 'background-color: #A5FF33;';
+      row.style = 'background-color: #DFFF00;';
     }
     else {
       row.style = 'background-color: white;';
