@@ -90,10 +90,6 @@ class HarmonyVisitorImpl(HarmonyVisitor):
             return bv[0]
         return bv
 
-    # Visit a parse tree produced by HarmonyParser#comp_op.
-    def visitComp_op(self, ctx: HarmonyParser.Comp_opContext):
-        return self.get_token(ctx.start, ctx.getText())
-
     # Visit a parse tree produced by HarmonyParser#unary_op.
     def visitUnary_op(self, ctx: HarmonyParser.Unary_opContext):
         return self.get_token(ctx.start, ctx.getText())
@@ -183,7 +179,7 @@ class HarmonyVisitorImpl(HarmonyVisitor):
 
     # Visit a parse tree produced by HarmonyParser#expr_stmt.
     def visitExpr_stmt(self, ctx: HarmonyParser.Expr_stmtContext):
-        e = self.visit(ctx.tuple_rule())
+        e = self.visit(ctx.expr_rule())
         tkn = self.get_token(ctx.start, ctx.start.text)
         endtoken = self.get_token(ctx.stop, ctx.stop.text)
         return CallAST(endtoken, tkn, False, e)
@@ -232,18 +228,18 @@ class HarmonyVisitorImpl(HarmonyVisitor):
     # Visit a parse tree produced by HarmonyParser#trap_stmt.
     def visitTrap_stmt(self, ctx: HarmonyParser.Trap_stmtContext):
         func = self.visit(ctx.expr())
-        if not isinstance(func, ApplyAST):
-            tkn = self.get_token(func.ast_token, func.ast_token[0])
-            raise HarmonyCompilerError(
-                message="Expected a method call but found something else.",
-                filename=self.file,
-                line=tkn[2],
-                column=tkn[3],
-                lexeme=tkn[0]
-            )
+        # if not isinstance(func, ApplyAST):
+        #     tkn = self.get_token(func.token, func.token[0])
+        #     raise HarmonyCompilerError(
+        #         message="Expected a method call but found something else.",
+        #         filename=self.file,
+        #         line=tkn[2],
+        #         column=tkn[3],
+        #         lexeme=tkn[0]
+        #     )
         tkn = self.get_token(ctx.start, ctx.start.text)
         endtoken = self.get_token(ctx.stop, ctx.stop.text)
-        return TrapAST(endtoken, tkn, False, func.method, func.arg)
+        return TrapAST(endtoken, tkn, False, func)
 
     # Visit a parse tree produced by HarmonyParser#pass_stmt.
     def visitPass_stmt(self, ctx: HarmonyParser.Pass_stmtContext):
@@ -294,19 +290,20 @@ class HarmonyVisitorImpl(HarmonyVisitor):
     def visitSpawn_stmt(self, ctx: HarmonyParser.Spawn_stmtContext):
         is_eternal = ctx.ETERNAL() is not None
         target = self.visit(ctx.expr())
-        if not isinstance(target, ApplyAST):
-            tkn = self.get_token(target.ast_token, target.ast_token[0])
-            raise HarmonyCompilerError(
-                message="Expected a method call but found something else.",
-                filename=self.file,
-                line=tkn[2],
-                column=tkn[3],
-                lexeme=tkn[0]
-            )
-        method, arg = target.method, target.arg
+        # if not isinstance(target, ApplyAST):
+        #     tkn = self.get_token(target.token, target.token[0])
+        #     raise HarmonyCompilerError(
+        #         message="Expected a method call but found something else.",
+        #         filename=self.file,
+        #         line=tkn[2],
+        #         column=tkn[3],
+        #         lexeme=tkn[0]
+        #     )
+        # method, arg = target.method, target.arg
         tkn = self.get_token(ctx.start, ctx.start.text)
         endtoken = self.get_token(ctx.stop, ctx.stop.text)
-        return SpawnAST(endtoken, tkn, False, method, arg, None, is_eternal)
+        # return SpawnAST(endtoken, tkn, False, method, arg, None, is_eternal)
+        return SpawnAST(endtoken, tkn, False, target, None, is_eternal)
 
     # Visit a parse tree produced by HarmonyParser#go_stmt.
     def visitGo_stmt(self, ctx: HarmonyParser.Go_stmtContext):
@@ -323,6 +320,13 @@ class HarmonyVisitorImpl(HarmonyVisitor):
         endtoken = self.get_token(ctx.stop, ctx.stop.text)
         return SequentialAST(endtoken, tkn, False, expr)
 
+    # Visit a parse tree produced by HarmonyParser#global_stmt.
+    def visitGlobal_stmt(self, ctx: HarmonyParser.Global_stmtContext):
+        expr = [self.visit(e) for e in ctx.expr()]
+        tkn = self.get_token(ctx.start, ctx.start.text)
+        endtoken = self.get_token(ctx.stop, ctx.stop.text)
+        return GlobalAST(endtoken, tkn, False, expr)
+
     # Visit a parse tree produced by HarmonyParser#builtin_stmt.
     def visitBuiltin_stmt(self, ctx: HarmonyParser.Builtin_stmtContext):
         name = str(ctx.NAME())
@@ -657,37 +661,156 @@ class HarmonyVisitorImpl(HarmonyVisitor):
             return values[0]
         return TupleAST(endtoken, values, tkn)
 
+    def factor_expr(self, ops, exprs):
+        if len(ops) == 0:
+            return exprs[0]
+        if len(ops) == 1:
+            return NaryAST(exprs[1].endtoken, exprs[0].token, ops[0], exprs)
+        found = False
+        for o in ops:
+            if o[0] in { '/', '//', '%', 'mod' }:
+                if found:
+                    raise HarmonyCompilerError(
+                        message="Expression too complicated: use parentheses",
+                        filename=self.file,
+                        line=token[2],
+                        column=token[3],
+                        lexeme=token[0]
+                    )
+                found = True
+            else:
+                assert o[0] == '*'
+        return NaryAST(exprs[-1].endtoken, exprs[0].token, ops[-1],
+            [ NaryAST(exprs[-2].endtoken, exprs[0].token, ops[0], exprs[:-1]), exprs[-1] ])
+
+    def simple_expr(self, ops, exprs):
+        simple_ops = [ o for o in ops if o[0] in { '+', '-' } ]
+        if len(simple_ops) == 0:
+            return self.factor_expr(ops, exprs)
+        sub_ops = []
+        sub_exprs = [ exprs[0] ]
+        sub_asts = []
+        for i in range(len(ops)):
+            if ops[i] in simple_ops:
+                sub_asts.append(self.factor_expr(sub_ops, sub_exprs))
+                sub_ops = []
+                sub_exprs = [ exprs[i+1] ]
+            else:
+                sub_ops.append(ops[i])
+                sub_exprs.append(exprs[i+1])
+        sub_asts.append(self.factor_expr(sub_ops, sub_exprs))
+        ast = sub_asts[0]
+        for i in range(len(simple_ops)):
+            ast = NaryAST(sub_asts[i+1].endtoken, sub_asts[0].token, simple_ops[i], [ ast, sub_asts[i+1] ])
+        return ast
+
+    def arith_expr(self, ops, exprs):
+        assert len(ops) + 1 == len(exprs)
+        if len(ops) == 0:
+            return exprs[0]
+        if len(ops) == 1:
+            return NaryAST(exprs[1].endtoken, exprs[0].token, ops[0], exprs)
+
+        # See if all are the same associative operator
+        if ops[0][0] in self.associative_operators and all(o[0] == ops[0][0] for o in ops):
+            return NaryAST(exprs[-1].endtoken, exprs[0].token, ops[0], exprs)
+
+        # See if it's a simple arithmetic expression
+        if all(o[0] in { '+', '-', '*', '/', '//', '%', 'mod' } for o in ops):
+            return self.simple_expr(ops, exprs)
+
+        raise HarmonyCompilerError(
+            message="Expression too complicated: use parentheses",
+            filename=self.file,
+            line=token[2],
+            column=token[3],
+            lexeme=token[0]
+        )
+
+    def cmp_expr(self, ops, exprs):
+        assert len(ops) + 1 == len(exprs)
+        if len(ops) == 0:
+            return exprs[0]
+        cmp_ops = [ o for o in ops if o[0] in { '==', '!=', '<', '<=', '>', '>=' } ]
+        if len(cmp_ops) == 0:
+            return self.arith_expr(ops, exprs)
+        sub_ops = []
+        sub_exprs = [ exprs[0] ]
+        sub_asts = []
+        for i in range(len(ops)):
+            if ops[i] in cmp_ops:
+                sub_asts.append(self.arith_expr(sub_ops, sub_exprs))
+                sub_ops = []
+                sub_exprs = [ exprs[i+1] ]
+            else:
+                sub_ops.append(ops[i])
+                sub_exprs.append(exprs[i+1])
+        sub_asts.append(self.arith_expr(sub_ops, sub_exprs))
+        return CmpAST(sub_asts[-1].endtoken, sub_asts[0].token, cmp_ops, sub_asts)
+
+    def bool_expr_helper(self, ops, exprs, operator):
+        op = None
+        sub_ops = []
+        sub_exprs = [ exprs[0] ]
+        sub_asts = []
+        for i in range(len(ops)):
+            if ops[i][0] == operator:
+                if op == None:
+                    op = ops[i]
+                sub_asts.append(self.cmp_expr(sub_ops, sub_exprs))
+                sub_ops = []
+                sub_exprs = [ exprs[i+1] ]
+            else:
+                sub_ops.append(ops[i])
+                sub_exprs.append(exprs[i+1])
+        sub_asts.append(self.cmp_expr(sub_ops, sub_exprs))
+        return NaryAST(sub_asts[-1].endtoken, sub_asts[0].token, op, sub_asts)
+
+    def bool_expr(self, ops, exprs):
+        assert len(ops) > 0
+        assert len(ops) + 1 == len(exprs)
+        bool_ops = [ o[0] for o in ops if o[0] in { 'or', 'and', '=>', '==', '=' } ]
+        if 'or' in bool_ops:
+            if len(set(bool_ops)) > 1:
+                assert False, bool_ops
+                raise HarmonyCompilerError(
+                    message="Boolean expression too complicated: use parentheses",
+                    filename=self.file,
+                    line=token[2],
+                    column=token[3],
+                    lexeme=token[0]
+                )
+            return self.bool_expr_helper(ops, exprs, 'or')
+        if 'and' in bool_ops:
+            if len(set(bool_ops)) > 1:
+                raise HarmonyCompilerError(
+                    message="Boolean expression too complicated: use parentheses",
+                    filename=self.file,
+                    line=token[2],
+                    column=token[3],
+                    lexeme=token[0]
+                )
+            return self.bool_expr_helper(ops, exprs, 'and')
+        if '=>' in bool_ops:
+            if len(bool_ops) > 1:
+                raise HarmonyCompilerError(
+                    message="Boolean imply expression too complicated: use parentheses",
+                    filename=self.file,
+                    line=token[2],
+                    column=token[3],
+                    lexeme=token[0]
+                )
+            return self.bool_expr_helper(ops, exprs, '=>')
+        return self.cmp_expr(ops, exprs)
+
     # Visit a parse tree produced by HarmonyParser#nary_expr.
     def visitNary_expr(self, ctx: HarmonyParser.Nary_exprContext):
         expressions = [self.visit(e) for e in ctx.expr_rule()]
         tkn = self.get_token(ctx.start, ctx.start.text)
         endtoken = self.get_token(ctx.stop, ctx.stop.text)
-        if ctx.comp_op():
-            comps = [self.visit(o) for o in ctx.comp_op()]
-            return CmpAST(endtoken, tkn, comps, expressions)
         if ctx.arith_op():
             ops = [self.visit(o) for o in ctx.arith_op()]
-            expected = ops[0]
-            if expected[0] not in self.associative_operators and len(ops) > 1:
-                second_op = ops[1]
-                raise HarmonyCompilerError(
-                    message=f"Cannot have expression with multiple non-associative operators.",
-                    filename=self.file,
-                    line=second_op[2],
-                    column=second_op[3],
-                    lexeme=second_op[0]
-                )
-            for actual in ops[1:]:
-                if actual[0] != expected[0]:
-                    raise HarmonyCompilerError(
-                        message=f"Cannot have multiple types of operators in one expression. Expected {expected[0]} but got {actual[0]}",
-                        filename=self.file,
-                        line=actual[2],
-                        column=actual[3],
-                        lexeme=actual[0]
-                    )
-            binop = ops[0]
-            return NaryAST(endtoken, tkn, binop, expressions)
+            return self.bool_expr(ops, expressions)
         if ctx.IF():
             condition = self.visit(ctx.nary_expr())
             expressions.insert(1, condition)
@@ -700,6 +823,7 @@ class HarmonyVisitorImpl(HarmonyVisitor):
                 not_token = self.get_token(ctx.NOT().symbol, str(ctx.NOT()))
                 return NaryAST(endtoken, tkn, not_token, [ast])
             return ast
+        assert len(expressions) == 1
         return expressions[0]
 
     # Visit a parse tree produced by HarmonyParser#expr_rule.
@@ -715,16 +839,17 @@ class HarmonyVisitorImpl(HarmonyVisitor):
         if ctx.STOP():
             expr = self.visit(ctx.expr_rule())
             return StopAST(endtoken, tkn, expr)
-        if ctx.POINTER_OF():
-            expr = self.visit(ctx.expr_rule())
-            return PointerAST(endtoken, expr, tkn)
-        if ctx.ADDRESS_OF():
-            expr = self.visit(ctx.expr_rule())
-            return AddressAST(endtoken, expr, tkn)
         if ctx.unary_op():
             op = self.visit(ctx.unary_op())
-            expr = self.visit(ctx.expr_rule())
-            return NaryAST(endtoken, tkn, op, [expr])
+            if op[0] == '!':
+                expr = self.visit(ctx.expr_rule())
+                return PointerAST(endtoken, expr, tkn)
+            elif op[0] == '?':
+                expr = self.visit(ctx.expr_rule())
+                return AddressAST(endtoken, expr, tkn)
+            else:
+                expr = self.visit(ctx.expr_rule())
+                return NaryAST(endtoken, tkn, op, [expr])
         if ctx.application():
             return self.visit(ctx.application())
         raise HarmonyCompilerError(
@@ -739,10 +864,16 @@ class HarmonyVisitorImpl(HarmonyVisitor):
     def visitApplication(self, ctx:HarmonyParser.ApplicationContext):
         tkn = self.get_token(ctx.start, ctx.start.text)
         endtoken = self.get_token(ctx.stop, ctx.stop.text)
-        if ctx.ARROW():
+        if ctx.ARROWID():
             p = self.visit(ctx.application())
-            name = self.get_token(ctx.NAME().symbol, str(ctx.NAME()))
-            return ApplyAST(endtoken, PointerAST(endtoken, p, tkn), ConstantAST(endtoken, name), tkn)
+            name = self.get_token(ctx.ARROWID().symbol, str(ctx.ARROWID()))
+            (lexeme, file, line, col) = name
+            lexeme = lexeme[2:]
+            col += 2
+            while lexeme[0] == ' ':
+                lexeme = lexeme[1:]
+                col += 1
+            return ApplyAST(endtoken, PointerAST(endtoken, p, tkn), ConstantAST(endtoken, (lexeme, file, line, col)), tkn)
         elif ctx.application():
             f = self.visit(ctx.application())
             return ApplyAST(endtoken, f, self.visit(ctx.basic_expr()), tkn)
